;; in JACL package; use JACL from COMMON-LISP and switch to COMMON-LISP
(%let ((cl-pkg (\. (%js "Package") (|get| (\. '#:common-lisp |name|)))))
  (\. cl-pkg (|usePackage| cl:*package*))
  (%setq cl:*package* cl-pkg))

(\. *package* (|exportSymbol| (\. 'defmacro |name|)))
(%js "(~{}.setMacro().fvalue = ~{})"
     'defmacro
     (%lambda (env form name params &rest body)
       ;; TODO Use %PROGN instead of %LET
       `(jacl:%progn
          (jacl:%js "(~{}.setMacro().fvalue = ~{})"
            ',name
            ;; TODO Macro lambda list parsing. Also, consider
            ;; incorportating a symbol's package into the name of its
            ;; generated local.
            (jacl:%lambda (jacl::&environment jacl::&whole ,@params)
              ,@body))
          ',name)))

(\. *package* (|exportSymbol| (\. 'defun |name|)))
(defmacro defun (name params &rest body)
  `(jacl:%progn
     (jacl:%js "(~{}.fvalue = ~{})"
       ',name
       (jacl:%lambda ,params ,@body))
     ',name))

(defun %export (symbol &optional (package *package*))
  (\. package (|exportSymbol| (\. symbol |name|))))

(%export 'let)
(defmacro let (bindings &rest body)
  `(jacl:%let ,bindings ,@body))

(%export 'if)
(defmacro if (test true-form &optional else-form)
  `(jacl:%if ,test ,true-form ,else-form))

(%export 'progn)
(defmacro progn (&rest forms)
  `(jacl:%progn ,@forms))

(%export 'when)
(defmacro when (test &rest forms)
  `(if ,test (progn ,@forms)))

(%export 'defvar)
(defmacro defvar (symbol &optional (value nil value?))
  `(progn
    (jacl:%js "~{}.isSpecial = true" ',symbol)
    (when (jacl:%js "(((~{}.value === UNDEFINED) && ~{}) ? true : null)" ',symbol ,value?)
      (jacl:%js "(~{}.value = ~{})" ',symbol ',value))
    ',symbol))

(%export 'eq)
(defun eq (x y)
  (jacl:%js "~{} === ~{} ? true : null" x y))

;; TODO Real EQL
(%export 'eql)
(defun eql (x y)
  (eq x y))

(%export 'null)
(defun null (x) (eq x nil))

(%export 'not)
(defun not (x) (eq x nil))

(%export 'cons)
(defun cons (car cdr)
  (jacl:%new (jacl:%js "Cons") car cdr))

(%export 'car)
(defun car (x)
  (when (not (null x))
    (jacl:\. x |car|)))

(%export 'cdr)
(defun cdr (x)
  (when (not (null x))
    (jacl:\. x |cdr|)))

(%export 'rplacd)
(defun rplacd (cons obj)
  ;; TODO type check
  (jacl:%js "~{}.cdr=~{}" cons obj)
  cons)

(%export 'cadr)
(defun cadr (x)
  (car (cdr x)))

(%export 'cddr)
(defun cddr (x)
  (cdr (cdr x)))

(%export 'caar)
(defun caar (x)
  (car (car x)))

(%export 'cadar)
(defun cadar (x)
  (car (cdr (car x))))

(%export 'cond)
(defmacro cond (&rest clauses)
  (when clauses
    `(if ,(caar clauses)
         ,(cadar clauses)
       (cond ,@(cdr clauses)))))

(%export 'numberp)
(defun numberp (x)
  (jacl:%js "typeof ~{} === 'number' ? true : null" x))

(%export 'stringp)
(defun stringp (x)
  (jacl:%js "~{} instanceof LispString ? true : null" x))

(%export 'symbolp)
(defun symbolp (x)
  (jacl:%js "~{} instanceof LispSymbol ? true : null" x))

(%export 'consp)
(defun consp (x)
  (jacl:%js "~{} instanceof Cons ? true : null" x))

(%export 'tagbody)
(defmacro tagbody (&rest body)
  `(jacl:%tagbody ,@body))

(%export 'go)
(defmacro go (tag)
  `(jacl:%go ,tag))

(%export '*gensym-counter*)
(defvar *gensym-counter* 0)

(defmacro %type-error (expected-type)
  `(jacl:%throw
    (jacl:%new (jacl:%js "TypeError")
               (jacl:%js "'Not a ' + ~{}.toString()" ,expected-type))))

(%export 'set)
(defun set (symbol value)
  (when (not (symbolp symbol))
    (%type-error "symbol"))
  (jacl:%js "(~{}.value = ~{})" symbol value))

(%export '1+)
(defun 1+ (x)
  (when (not (numberp x)) (%type-error "number"))
  (jacl:%js "~{}+1" x))

(%export '1-)
(defun 1- (x)
  (when (not (numberp x)) (%type-error "number"))
  (jacl:%js "~{}-1" x))

(defun %aref (arr i)
  (jacl:%js "~{}[~{}]" arr i))

;(%export '+)
;(defun + (&rest nums)
;  (declare (jacl:rest-array))
;  (let ((len (\. nums |length|))
;        (sum 0)
;        (i 0))
;    (tagbody
;     start
;     (when (< i len)
;       (setq sum (jacl:%js "~{}+~{}" sum (%aref nums i)))
;       (setq i (1+ i))
;       (go start)))
;    sum))

(%export 'gensym)
(defun gensym (&optional (x nil x?))
  (cond ((and x? (stringp x))
         (jacl:%new (jacl:%js "LispSymbol")
                    (jacl:%js "~{}+~{}.toString()" x *gensym-counter*)))
        ((and x? (numberp x))
         (jacl:%new (jacl:%js "LispSymbol")
                    (jacl:%js "'G'+~{}.toString()" x)))
        (x? (%type-error "string or integer"))
        (t
         (let ((sym (jacl:%new (jacl:%js "LispSymbol")
                               (jacl:%js "'G'+~{}.toString()" *gensym-counter*))))
           (set '*gensym-counter* (1+ *gensym-counter*))
           sym))))

(%export 'or)
(defmacro or (&rest forms)
  (when forms
    (let ((x (gensym)))
      `(let ((,x ,(car forms)))
         (if ,x ,x (or ,@(cdr forms)))))))

(%export 'and)
(defmacro and (&rest forms)
  (if forms
      (let ((x (gensym)))
        `(let ((,x ,(car forms)))
           (when ,x (and ,@(cdr forms)))))
      t))

(%export 'let*)
(defmacro let* (bindings &rest body)
  (if bindings
      `(let (,(car bindings))
         (let* ,(cdr bindings) ,@body))
      `(progn ,@body)))

(%export 'listp)
(defun listp (x)
  (or (null x) (consp x)))

(%export 'lambda)
(defmacro lambda (params &rest body)
  `(jacl:%lambda ,params ,@body))

(%export 'list)
(defun list (&rest objects)
  `(,@objects))

(%export 'functionp)
(defun functionp (x)
  (jacl:%js "~{} instanceof Function ? true : null" x))

(%export 'funcall)
(defun funcall (f &rest args)
  (declare (jacl:rest-array))
  (let ((func (cond ((symbolp f) (\. f (|func|)))
                    ((functionp f) f)
                    (t (%type-error "function or symbol")))))
    (\. func (|apply| nil args))))

(defun %map-pairs (fun pairs)
  (let ((head nil)
        (tail nil))
    (tagbody
     start
     (when (not pairs) (go end))
     (when (not head)
       (jacl:%setq head (list (funcall fun (car pairs) (cadr pairs))))
       (jacl:%setq tail head)
       (go next))
     (let ((new-tail (list (funcall fun (car pairs) (cadr pairs)))))
       (rplacd tail new-tail)
       (jacl:%setq tail new-tail))
     next
     (jacl:%setq pairs (cddr pairs))
     (go start)
     end)
    head))
;
(%export 'setq)
(defmacro setq (&rest pairs)
  `(cl:progn
    ,@(%map-pairs (lambda (x y) `(jacl:%setq ,x ,y)) pairs)))

(%export 'zerop)
(defun zerop (x)
  (when (not (numberp x)) (%type-error "number"))
  (eql x 0))

(defun %> (x y)
  (when (not (numberp x)) (%type-error "number"))
  (when (not (numberp y)) (%type-error "number"))
  (jacl:%js "~{}>~{}?true:null" x y))

(defun %< (x y)
  (when (not (numberp x)) (%type-error "number"))
  (when (not (numberp y)) (%type-error "number"))
  (jacl:%js "~{}<~{}?true:null" x y))

(%export '>)
(defun > (number &rest more-numbers)
  (when (not (numberp number))
    (%type-error "number"))
  (let ((ret t))
    (tagbody
     start
     (when (not more-numbers) (go end))
     (when (not (%> number (car more-numbers)))
       (setq ret nil)
       (go end))
     (setq number (car more-numbers)
           more-numbers (cdr more-numbers))
     (when (not (numberp number))
       (%type-error "number"))
     (go start)
     end)
    ret))

(%export '<)
(defun < (num &rest nums)
  (declare (jacl:rest-array))
  (let ((len (jacl:%dot nums |length|))
        (ret t)
        (i 0))
    (tagbody
     start
     (when (not (numberp num))
       (%type-error "number"))
     (when (eql i len) 
       (go end))
     (let ((x (%aref nums i)))
       (when (not (%< num x)) 
         (setq ret nil)
         (go end))
       (setq num x
             i (1+ i)))
     (go start)
     end)
    ret))

;;(defun %designated-string (x)
;;  (cond ((stringp x) x)
;;        ((symbolp x)
;;         (jacl:\. (jacl:%js "LispString") (|fromString| (jacl:\. x |name|))))
;;        (t (%type-error "string or symbol"))))
;;
;;(defun %designated-symbols (x)
;;  (cond ((symbolp x) (list x))
;;        ((listp x) x)
;;        (t (%type-error "symbol or list of symbols"))))
;;
;;(%export 'let*)
;;(defmacro let* (bindings &rest body)
;;  (if bindings
;;      `(let (,(car bindings))
;;         (let* ,(cdr bindings)
;;           ,@body))
;;      `(progn ,@body)))
;;
;;(%export 'function)
;;(defmacro function (x)
;;  `(jacl:%js "~{}.func()" x))
;;
;;(%export 'functionp)
;;(defun functionp (x)
;;  (jacl:%js "~{} instanceof Function ? true : null" x))
;;
;;(%export 'funcall)
;;(defun funcall (f &rest args)
;;  (when (not (functionp f))
;;    (%type-error "function"))
;;  (jacl:%js "~{}.call(null, List.toArray(~{}))" f args))
;;
;;(%export 'export)
;;(defun export (symbols &optional (package *package*))
;;  (let ((syms (designated-symbols symbols)))
;;    (tagbody
;;     start
;;     (when syms
;;       (%export (car syms) package)
;;       (jacl:%setq syms (cdr syms))
;;       (go start)))))

;; Use COMMON-LISP from COMMON-LISP-USER and switch to
;; COMMON-LISP-USER
(%let ((cl-user-pkg (\. (%js "Package") (|get| (\. '#:common-lisp-user |name|)))))
  (\. cl-user-pkg (|usePackage| cl:*package*))
  (%setq cl:*package* cl-user-pkg))
